@@ -72,7 +72,7 @@ class Prompt:
         display_out = re.sub(r"__GMARKER_START_variable_ref\$([^\$]*)\$___", r"<span style='background-color: rgb(0, 138.56128016, 250.76166089, 0.25); display: inline;' title='\1'>", display_out)
         display_out = display_out.replace("__GMARKER_END_variable_ref$$___", "</span>")
         display_out = display_out.replace("__GMARKER_each$$___", "<div style='border-left: 1px dashed rgb(0, 0, 0, .2); border-top: 0px solid rgb(0, 0, 0, .2); margin-right: -4px; display: inline; width: 4px; height: 24px;'></div>")
-        display_out = re.sub(r"__GMARKER_START_([^\$]*)\$([^\$]*)\$___", r"<span style='background-color: rgb(0, 165, 0, 0.25); display: inline;' title='\2'>", display_out)
+        display_out = re.sub(r"__GMARKER_START_([^\$]*)\$([^\$]*)\$___", r"<span style='background-color: rgb(165, 165, 165, 0.25); display: inline;' title='\2'>", display_out)
         display_out = re.sub(r"__GMARKER_END_([^\$]*)\$\$___", "</span>", display_out)
         display_out = "<pre style='padding: 7px; border-radius: 4px; background: white; white-space: pre-wrap; font-family: ColfaxAI, Arial; font-size: 16px; line-height: 24px; color: #000'>"+display_out+"</pre>"
 
@@ -132,7 +132,7 @@ class TopDownVisitor():
         self.variable_stack = [variables]
         self.prompt_object = prompt_object
     
-    def visit(self, node):
+    def visit(self, node, next_node=None):
 
         if node.expr_name == 'variable_ref':
             return self.get_variable(node.text)
@@ -171,7 +171,7 @@ class TopDownVisitor():
             return ast.literal_eval(node.text)
 
         elif node.expr_name == 'command':
-            visited_children = [str(self.visit(child)) for child in node.children]
+            visited_children = [str(self.visit(child, next_node)) for child in node.children]
             out = "".join(visited_children) or ""
             
             command_head = node.children[1].children[0]
@@ -216,6 +216,11 @@ class TopDownVisitor():
                     named_args["parser"] = self
                 if "partial_output" in sig.parameters:
                     named_args["partial_output"] = self._extend_prefix
+                if "next_text" in sig.parameters:
+                    if next_node is not None:
+                        named_args["next_text"] = next_node.text
+                    else:
+                        named_args["next_text"] = ""
 
                 command_output = command_function(*positional_args, **named_args)
 
@@ -255,11 +260,11 @@ class TopDownVisitor():
                     named_args["parser"] = self
                 if "block_content" in sig.parameters:
                     block_content = [node.children[1]]
-                    for child in node.children[2:-1]:
+                    for child in node.children[2].children:
                         if child.text == '':
                             continue
-                        block_content.append(child.children[0].children[0])
-                        block_content.append(child.children[0].children[1])
+                        block_content.append(child.children[0])
+                        block_content.append(child.children[1])
                     named_args["block_content"] = block_content
                 if "partial_output" in sig.parameters:
                     named_args["partial_output"] = self._extend_prefix
@@ -273,7 +278,14 @@ class TopDownVisitor():
             # end_block = self.visit(node.children[2])
 
         else:
-            visited_children = [self.visit(child) for child in node.children]
+            visited_children = []
+            for i, child in enumerate(node.children):
+                if len(node.children) > i + 1:
+                    inner_next_node = node.children[i + 1]
+                else:
+                    inner_next_node = next_node
+                visited_children.append(self.visit(child, inner_next_node))
+            # visited_children = [self.visit(child) for child in node.children]
             
             if len(visited_children) == 1:
                 return visited_children[0]
@@ -281,6 +293,11 @@ class TopDownVisitor():
                 return "".join(visited_children) or ""
 
     def get_variable(self, name, default_value=""):
+        if name == "True":
+            return True
+        elif name == "False":
+            return False
+        
         parts = name.split(".")
         for variables in reversed(self.variable_stack):
             curr_pos = variables
@@ -325,7 +342,7 @@ class TopDownVisitor():
             if found:
                 break
         if not found:
-            assert len(parts) == 1, "Can't set a property of a non-existing variable"
+            assert len(parts) == 1, "Can't set a property of a non-existing variable: " + name
             self.variable_stack[0][name] = value
 
     def _extend_prefix(self, text):
@@ -336,15 +353,23 @@ class TopDownVisitor():
             sys.stdout.flush()
 
 
-def _generate(variable_name, partial_output, stop=None, max_tokens=500, temperature=0.0, top_p=1.0, parser_prefix=None, parser=None, prefix="", suffix=""):
+def _generate(variable_name, partial_output, parse=False, stop=None, max_tokens=500, temperature=0.0, top_p=1.0, parser_prefix=None, parser=None, prefix="", suffix="", next_text=None):
     ''' Use the LM to generate a completion string that is stored in the variable `variable_name`.
     '''
+
+    # if stop is None then we use the text of the node after the generate command
+    if stop is None:
+        stop = next_text
+    
     gen_obj = parser.prompt_object.generator(parser_prefix+prefix, stop=stop, max_tokens=max_tokens, temperature=temperature, top_p=top_p)
     generated_value = prefix+gen_obj["choices"][0]["text"]+suffix
     parser.set_variable(variable_name, generated_value)
-    subtree = grammar.parse(generated_value)
-    parsed_text = parser.visit(subtree)
-    return parsed_text
+    if parse:
+        subtree = grammar.parse(generated_value)
+        return parser.visit(subtree)
+    else:
+        partial_output(generated_value)
+        return generated_value
 
 def _add(*args):
     ''' Add the given variables together.
@@ -356,7 +381,7 @@ def _subtract(arg1, arg2):
     '''
     return arg1 - arg2
 
-def _each(list, block_content, parser, parser_prefix=None, stop=None):
+def _each(list, block_content, parser, parser_prefix=None, stop=None, batch_generate=False, batch_generate_temperature=0.0, batch_generate_max_tokens=500, batch_generate_top_p=1.0):
     ''' Iterate over a list and execute a block for each item.
     '''
     assert len(block_content) == 1
@@ -367,21 +392,52 @@ def _each(list, block_content, parser, parser_prefix=None, stop=None):
     if isinstance(list, str):
         assert stop is not None, "Must provide a stop token when doing variable length iteration!"
         stop_tokens = parser.prompt_object.generator.tokenize(stop)
-        i = 0
-        data = []
-        while True:
-            parser.variable_stack.append({})
-            parser.variable_stack[-1]["@index"] = i
-            parser.variable_stack[-1]["@first"] = i == 0
-            parser.variable_stack[-1]["this"] = {}
-            out.append(parser.visit(block_content[0]))
-            data.append(parser.variable_stack.pop()["this"])
-            i += 1
 
-            # we run a quick generation to see if we have reached the end of the list (not the +2 tokens is to help be tolorant to whitespace)
-            gen_obj = parser.prompt_object.generator(parser.prefix, stop=stop, max_tokens=len(stop_tokens)+2, temperature=0)
-            if gen_obj["choices"][0]["finish_reason"] == "stop":
-                break
+        if not batch_generate:
+            i = 0
+            data = []
+            while True:
+                parser.variable_stack.append({})
+                parser.variable_stack[-1]["@index"] = i
+                parser.variable_stack[-1]["@first"] = i == 0
+                parser.variable_stack[-1]["this"] = {}
+                out.append(parser.visit(block_content[0]))
+                data.append(parser.variable_stack.pop()["this"])
+                i += 1
+
+                # we run a quick generation to see if we have reached the end of the list (not the +2 tokens is to help be tolorant to whitespace)
+                gen_obj = parser.prompt_object.generator(parser.prefix, stop=stop, max_tokens=len(stop_tokens)+2, temperature=0)
+                if gen_obj["choices"][0]["finish_reason"] == "stop":
+                    break
+        else:
+            # create a pattern to match each item
+            pattern = re.sub(
+                r'{{generate [\'"]([^\'"]+)[\'"][^}]*}}',
+                lambda x: r"(?P<"+x.group(1).replace("this.", "")+">.*?)",
+                block_content[0].text
+            )
+
+            # generate the looped content
+            gen_obj = parser.prompt_object.generator(parser_prefix, stop=stop, max_tokens=batch_generate_max_tokens, temperature=batch_generate_temperature, top_p=batch_generate_top_p)
+            generated_value = gen_obj["choices"][0]["text"]
+
+            # parse the generated content (this assumes the generated content is syntactically correct)
+            matches = re.finditer(pattern, generated_value)
+            data = []
+            for m in matches:#f"__GMARKER_START_{name}${node_text}$___{out}__GMARKER_END_{name}$$___"
+                
+                # get the variables that were generated
+                match_dict = m.groupdict()
+                data.append(match_dict)
+
+                # recreate the output string with format markers added
+                item_out = re.sub(
+                    r"{{generate [\'\"]([^\'\"]+)[\'\"][^}]*}}",
+                    lambda x: "__GMARKER_START_generate$"+x.group()+"$___"+match_dict[x.group(1).replace("this.", "")]+"__GMARKER_END_generate$$___",
+                    block_content[0].text
+                )
+                out.append(item_out)
+                
         parser.set_variable(list, data)
 
     # if the list is not a string then it is a list of items to iterate over
@@ -419,7 +475,7 @@ def _select(variable_name, block_content, parser, partial_output, parser_prefix=
         logprobs=10
     )
 
-    # copmute logprobs for each option
+    # compute logprobs for each option
     top_logprobs = gen_obj["choices"][0]["logprobs"]["top_logprobs"]
     for i in range(len(top_logprobs)):
         for option in option_tokens:
@@ -437,7 +493,7 @@ def _select(variable_name, block_content, parser, partial_output, parser_prefix=
         parser.set_variable(logprobs, option_logprobs)
     
     if max(option_logprobs.values()) <= -100:
-        raise ValueError("No valid option generated in #select, this could be fixed if we used a tokenizer and forced the LM to use a valid option!")
+        raise ValueError("No valid option generated in #select, this could be fixed if we used a tokenizer and forced the LM to use a valid option! The top logprobs were" + str(top_logprobs))
     
     partial_output(selected_option)
 
